import os
import numpy as np
import cv2
from skimage import transform
import rasterio
from PIL import Image

def geometric_filtering(features_i, features_j, matches_ij, ransac_thr=0.3):
    """
    Given a series of pairwise matches, use OpenCV to fit a fundamental matrix using RANSAC to filter outliers
    The 7-point algorithm is used to derive the fundamental matrix
    https://docs.opencv.org/3.0-beta/modules/calib3d/doc/camera_calibration_and_3d_reconstruction.html#findfundamentalmat
    Args:
        features_i: N[i]x132 array representing the N[i] keypoints from image i
        features_j: N[j]x132 array representing the N[j] keypoints from image j
        matches_ij: Mx2 array representing M matches between features_i and features_j
        ransac_thr (optional): float, RANSAC outlier rejection threshold
    Returns:
        matches_ij: filtered version of matches_ij (will contain same amount of rows or less)
    """
    kp_coords_i = features_i[matches_ij[:, 0], :2]
    kp_coords_j = features_j[matches_ij[:, 1], :2]
    if ransac_thr is None:
        F, mask = cv2.findFundamentalMat(kp_coords_i, kp_coords_j, cv2.FM_RANSAC)
    else:
        F, mask = cv2.findFundamentalMat(kp_coords_i, kp_coords_j, cv2.FM_RANSAC, ransac_thr)

    #mask = inliers_mask_from_fundamental_matrix(F, kp_coords_i, kp_coords_j, ransac_thr)
    matches_ij = matches_ij[mask.ravel().astype(bool), :] if mask is not None else None
    return matches_ij

def opencv_match_SIFT(features_i, features_j, dst_thr=0.8, ransac_thr=0.3, matcher="bruteforce"):
    """
    Match SIFT keypoints using OpenCV matchers
    Args:
        features_i: N[i]x132 array representing the N[i] keypoints from image i
        features_j: N[j]x132 array representing the N[j] keypoints from image j
        dst_thr (optional): float, threshold for SIFT distance ratio test
        ransac_thr (optional): float, threshold for RANSAC geometric filtering using the fundamental matrix
        matcher (optional): string, identifies the OpenCV matcher to use: either "flann" or "bruteforce"
    Returns:
        matches_ij: Mx2 array representing M matches. Each match is represented by two values (i, j)
                    which means that the i-th kp/row in s2p_features_i matches the j-th kp/row in s2p_features_j
        n_matches_after_ratio_test: integer, the number of matches left after the SIFT distance ratio test
        n_matches_after_geofilt: integer, the number of matches left after RANSAC filtering
    """

    descriptors_i = features_i[:, 4:].astype(np.float32)
    descriptors_j = features_j[:, 4:].astype(np.float32)
    if matcher == "bruteforce":
        # Bruteforce matcher
        bf = cv2.BFMatcher()
        matches = bf.knnMatch(descriptors_i, descriptors_j, k=2)
    elif matcher == "flann":
        # FLANN matcher
        # from https://docs.opencv.org/3.4/dc/dc3/tutorial_py_matcher.html
        FLANN_INDEX_KDTREE = 1
        index_params = dict(algorithm=FLANN_INDEX_KDTREE, trees=5)
        search_params = dict(checks=50)  # or pass empty dictionary
        flann = cv2.FlannBasedMatcher(index_params, search_params)
        matches = flann.knnMatch(descriptors_i, descriptors_j, k=2)
    else:
        raise ValueError('ERROR: OpenCV matcher is not recognized ! Valid values are "flann" or "bruteforce"')

    # Apply ratio test as in Lowe's paper
    matches_ij = np.array([[m.queryIdx, m.trainIdx] for m, n in matches if m.distance < dst_thr * n.distance])
    n_matches_after_ratio_test = matches_ij.shape[0]

    # Geometric filtering using the Fundamental matrix
    if n_matches_after_ratio_test > 0:
        matches_ij = geometric_filtering(features_i, features_j, matches_ij, ransac_thr)
    else:
        # no matches were left after ratio test
        matches_ij = None
    n_matches_after_geofilt = 0 if matches_ij is None else matches_ij.shape[0]

    return matches_ij, n_matches_after_ratio_test, n_matches_after_geofilt

def opencv_detect_SIFT(input_image, input_mask=None):
    """
    Detect SIFT keypoints in a single input grayscale image using OpenCV
    Requirement: pip3 install opencv-contrib-python==3.4.0.12
    Documentation of opencv keypoint class: https://docs.opencv.org/3.4/d2/d29/classcv_1_1KeyPoint.html
    Args:
        input_image: numpy 2d array
        input_mask: numpy binary mask, to restrict the search of keypoints to a certain area,
                    parts of the mask with 0s are not explored
    Returns:
        features: Nx132 array, where N is the number of SIFT keypoints detected in image i
                  each row/keypoint is represented by 132 values:
                  (col, row, scale, orientation) in columns 0-3 and (sift_descriptor) in the following 128 columns
        n_kp: integer, number of keypoints detected
    """
    input_image = input_image.astype(np.uint8)
    input_mask = None if input_mask is None else input_mask.astype(np.uint8)

    sift = cv2.SIFT_create() # cv2.xfeatures2d.SIFT_create() for older opencv versions
    kp, des = sift.detectAndCompute(input_image, input_mask)
    features = np.array([[*k.pt, k.size, k.angle, *d] for k, d in zip(kp, des)])

    return features

def compute_matches(img_ref, img_sec):
    features_ref = opencv_detect_SIFT(img_ref)
    features_sec = opencv_detect_SIFT(img_sec)
    matches_idx, _, _ = opencv_match_SIFT(features_ref, features_sec, 0.6)
    matches = np.hstack([features_ref[matches_idx[:, 0], :2], features_sec[matches_idx[:, 1], :2]])
    return matches


def affine_fundamental_matrix(matches):
    """
    Estimates the affine fundamental matrix given a set of point correspondences
    between two images.
    Arguments:
        matches: 2D array of size Nx4 containing a list of pairs of matching
            points. Each line is of the form x1, y1, x2, y2, where (x1, y1) is
            the point in the first view while (x2, y2) is the matching point in
            the second view.
    Returns:
        the estimated affine fundamental matrix, given by the Gold Standard
        algorithm, as described in Hartley & Zisserman book (see chap. 14).
    """
    # revert the order of points to fit H&Z convention (see algo 14.1)
    X = matches[:, [2, 3, 0, 1]]

    # compute the centroid
    N = len(X)
    XX = np.sum(X, axis=0) / N

    # compute the Nx4 matrix A
    A = X - np.tile(XX, (N, 1))

    # the solution is obtained as the singular vector corresponding to the
    # smallest singular value of matrix A. See Hartley and Zissermann for
    # details.
    # It is the last line of matrix V (because np.linalg.svd returns V^T)
    U, S, V = np.linalg.svd(A)
    N = V[-1, :]

    # extract values and build F
    F = np.zeros((3, 3))
    F[0, 2] = N[0]
    F[1, 2] = N[1]
    F[2, 0] = N[2]
    F[2, 1] = N[3]
    F[2, 2] = -np.dot(N, XX)

    return F

def rectifying_similarities_from_affine_fundamental_matrix(F):
    """
    Computes two similarities from an affine fundamental matrix.
    Args:
        F: 3x3 numpy array representing the input fundamental matrix
        debug (optional, default is False): boolean flag to activate verbose
            mode
    Returns:
        S, S': two similarities such that, when used to resample the two images
            related by the fundamental matrix, the resampled images are
            stereo-rectified.
    """
    # check that the input matrix is an affine fundamental matrix
    assert(np.shape(F) == (3, 3))
    assert(np.linalg.matrix_rank(F) == 2)
    np.testing.assert_allclose(F[:2, :2], np.zeros((2, 2)))

    # notations
    a = F[0, 2]
    b = F[1, 2]
    c = F[2, 0]
    d = F[2, 1]
    e = F[2, 2]

    # rotations
    r = np.sqrt(c*c + d*d)
    s = np.sqrt(a*a + b*b)
    R1 = 1 / r * np.array([[d, -c], [c, d]])
    R2 = 1 / s * np.array([[-b, a], [-a, -b]])

    # zoom and translation
    z = np.sqrt(r / s)
    t = 0.5 * e / np.sqrt(r * s)

    # output similarities
    S1 = np.zeros((3, 3))
    S1[0:2, 0:2] = z * R1
    S1[1, 2] = t
    S1[2, 2] = 1

    S2 = np.zeros((3, 3))
    S2[0:2, 0:2] = 1 / z * R2
    S2[1, 2] = -t
    S2[2, 2] = 1

    return S1, S2

def points_apply_homography(H, pts):
    """
    Applies an homography to a list of 2D points.
    Args:
        H: numpy array containing the 3x3 homography matrix
        pts: numpy array containing the list of 2D points, one per line
    Returns:
        a numpy array containing the list of transformed points, one per line
    """
    # if the list of points is not a numpy array, convert it
    if (type(pts) == list):
        pts = np.array(pts)

    # convert the input points to homogeneous coordinates
    if len(pts[0]) < 2:
        raise ValueError(
            "The input must be a numpy array"
            "of 2D points, one point per line"
        )
    pts = np.hstack((pts[:, 0:2], pts[:, 0:1]*0+1))

    # apply the transformation
    Hpts = (np.dot(H, pts.T)).T

    # normalize the homogeneous result and trim the extra dimension
    Hpts = Hpts * (1.0 / np.tile( Hpts[:, 2], (3, 1)) ).T
    return Hpts[:, 0:2]

def bounding_box2D(pts):
    """
    bounding box for the points pts
    """
    dim = len(pts[0])  # should be 2
    bb_min = [min([t[i] for t in pts]) for i in range(dim)]
    bb_max = [max([t[i] for t in pts]) for i in range(dim)]
    return bb_min[0], bb_min[1], bb_max[0] - bb_min[0], bb_max[1] - bb_min[1]

def matrix_translation(x, y):
    t = np.eye(3)
    t[0, 2] = x
    t[1, 2] = y
    return t

def rectification_homographies(matches, x, y, w, h):
    """
    Computes rectifying homographies from point matches for a given ROI.
    The affine fundamental matrix F is estimated with the gold-standard
    algorithm, then two rectifying similarities (rotation, zoom, translation)
    are computed directly from F.
    Args:
        matches: numpy array of shape (n, 4) containing a list of 2D point
            correspondences between the two images.
        x, y, w, h: four integers defining the rectangular ROI in the first
            image. (x, y) is the top-left corner, and (w, h) are the dimensions
            of the rectangle.
    Returns:
        S1, S2, F: three numpy arrays of shape (3, 3) representing the
        two rectifying similarities to be applied to the two images and the
        corresponding affine fundamental matrix.
    """
    # estimate the affine fundamental matrix with the Gold standard algorithm
    F = affine_fundamental_matrix(matches)

    # compute rectifying similarities
    S1, S2 = rectifying_similarities_from_affine_fundamental_matrix(F)

    # pull back top-left corner of the ROI to the origin (plus margin)
    pts = points_apply_homography(S1, [[x, y], [x+w, y], [x+w, y+h], [x, y+h]])
    x0, y0 = bounding_box2D(pts)[:2]
    T = matrix_translation(-x0, -y0)
    return np.dot(T, S1), np.dot(T, S2), F

def image_apply_homography(input_image, H, h, w):
    H_ = transform.ProjectiveTransform(matrix=H)
    output_image = transform.warp(input_image, H_.inverse, output_shape=(h, w),
                                  cval=np.nan, mode='constant', preserve_range=True, order=5)
    return output_image

def pts_apply_homography(pts2d, H):
    tform = transform.ProjectiveTransform(matrix=H)
    return tform(pts2d)

def read_input_image(img_path):
    if img_path.endswith(".tif"):
        with rasterio.open(img_path, "r") as f:
            img = np.transpose(f.read(), (1, 2, 0))
    else:
        img = np.array(Image.open(img_path))
    return img.astype(np.float32)

def save_png(img_path, img):
    im = Image.fromarray(np.clip(np.uint8(img), 0, 255))
    im.save(img_path)

def save_tif(img_path, img):
    if len(img.shape) == 2:
        img = img[..., np.newaxis]
    # expected image shape is H*W*C
    profile = {}
    profile["dtype"] = img.dtype
    profile["height"] = img.shape[0]
    profile["width"] = img.shape[1]
    profile["count"] = img.shape[2]
    with rasterio.open(img_path, "w", **profile) as f:
        f.write(np.transpose(img, (2, 0, 1)))

def refine_rectiying_homographies(matches, H1, H2):
    # this function refines the rectifying homographies H1 and H2 to ensure unidirectional movement to the left

    pts1 = pts_apply_homography(matches[:, :2], H1)
    pts2 = pts_apply_homography(matches[:, 2:], H2)

    def func(p, x, y):
        y_ = p[0] * x[:, 0] + p[1] * x[:, 1] + p[2]
        return abs(y[:, 0] - y_)
    from scipy import optimize
    p0 = np.array([0., 0., 0.])
    popt = optimize.leastsq(func, p0, args=(pts1, pts2))[0]
    a, b, c = popt
    H1 = np.array([[a, b, c], [0, 1, 0], [0, 0, 1]]) @ H1

    pts1 = pts_apply_homography(matches[:, :2], H1)
    pts2 = pts_apply_homography(matches[:, 2:], H2)
    disp_ = pts1 - pts2
    x0 = disp_[:, 0].min() # disp_[:, 0] must be positive in all matches to have all displacements in left direction
    margin = 20 # margin in pixels
    if x0 < 0:
        H1 = matrix_translation(-x0+margin, 0.) @ H1

    return H1, H2

def get_file_id(filename):
    """
    return what is left after removing directory and extension from a path
    """
    return os.path.splitext(os.path.basename(filename))[0]

def rectify_pair_custom(ref_path, sec_path, x, y, w, h, out_dir):
    # read input images
    img_ref = read_input_image(ref_path)
    img_sec = read_input_image(sec_path)
    # find sift matches
    matches = compute_matches(img_ref, img_sec)
    # compute rectifying homographies
    H1, H2, F = rectification_homographies(matches, x, y, w, h)
    # refine homographies and rectify
    H1, H2 = refine_rectiying_homographies(matches, H1, H2)
    rectified_ref = image_apply_homography(img_ref, H1, h, w)
    rectified_sec = image_apply_homography(img_sec, H2, h, w)
    # save intermediate files
    input_pair_id = get_file_id(ref_path)
    os.makedirs(os.path.join(out_dir, "rectified_left"), exist_ok=True)
    os.makedirs(os.path.join(out_dir, "rectified_right"), exist_ok=True)
    save_tif(os.path.join(out_dir, f'rectified_left/{input_pair_id}.tif'), rectified_ref)
    save_tif(os.path.join(out_dir, f'rectified_right/{input_pair_id}.tif'), rectified_sec)
    os.makedirs(os.path.join(out_dir, "H_left"), exist_ok=True)
    os.makedirs(os.path.join(out_dir, "H_right"), exist_ok=True)
    np.savetxt(os.path.join(out_dir, f'H_left/{input_pair_id}.txt'), H1, fmt='%12.6f')
    np.savetxt(os.path.join(out_dir, f'H_right/{input_pair_id}.txt'), H2, fmt='%12.6f')

