# from https://github.com/gfacciol/IS18/blob/master/stereo.py

import numpy as np
from numba import jit

def specklefilter(off, area=25, th=0):
    '''
    speckle filter of dispairt map off

    Args:
        off:  numpy array with the input disparity map
        area: the surface (in pixels) of the smallest allowed connected component of disparity
        th:   similarity threshold used to determin if two neighboring pixels have the same value

    Returns:
       numpy array with the filtered disparity map, removed points are set to infinity
    '''

    @jit
    def find(i, idx):  # finds the root of a dsf
        if idx.flat[i] == i:
            return i
        else:
            ret = find(idx.flat[i], idx)
            # idx.flat[i] = ret    // path compression is useles with idx passed by value
            return ret

    @jit
    def dsf(D, th=0):  # builds a dsf
        h, w = D.shape[0], D.shape[1]
        idx = np.zeros([h, w], dtype=int)
        for j in range(h):
            for i in range(w):
                idx[j, i] = j * w + i

        for j in range(h):
            for i in range(w):
                if (i > 0):
                    if (abs(D[j, i] - D[j, i - 1]) <= th):
                        a = find(idx[j, i], idx)
                        b = find(idx[j, i - 1], idx)
                        idx[j, i] = idx[j, i - 1]
                        idx.flat[a] = b

                if (j > 0):
                    if (abs(D[j, i] - D[j - 1, i]) <= th):
                        a = find(idx[j, i], idx)
                        b = find(idx[j - 1, i], idx)
                        idx[j, i] = idx[j - 1, i]
                        idx.flat[a] = b

        return idx

    @jit
    def labels(idx):
        h, w = idx.shape[0], idx.shape[1]
        lab = idx * 0

        for i in range(h * w):
            ind = find(i, idx)
            lab.flat[i] = ind
        return lab

    @jit
    def areas(lab):
        h, w = lab.shape[0], lab.shape[1]
        area = np.zeros(lab.shape, dtype=int)
        LL = np.zeros(lab.shape, dtype=int)
        for i in range(w * h):
            area.flat[lab.flat[i]] += 1
        for i in range(w * h):
            LL.flat[i] = area.flat[lab.flat[i]]
        return LL

    # build the dsf
    ind = dsf(off, th=th)
    # extract the labels of all the regions
    lab = labels(ind)
    # create map where all the regions are tagged with their area
    are = areas(lab)
    # filter the disparity map
    filtered = np.where((are > area), off, np.inf)

    return filtered

def leftright(offL, offR, maxdiff=1):
    '''
    Filters the disparity maps applying the left-right consistency test
        | offR(round(x - offL(x))) + offR(x)| <= maxdiff
    Args:
        offL, offR: numpy arrays containing the Left and Right disparity maps
        maxdiff: threshold for the uniqueness constraint
    Returns:
        numpy array containing the offL disparity map,
        where the rejected pixels are set to np.inf
    '''
    sh = offL.shape
    X, Y = np.meshgrid(range(sh[1]), range(sh[0]))
    X = np.minimum(np.maximum(X - offL.astype(int), 0), sh[1]-1)
    m = np.abs(offL - offR[Y,X] ) > maxdiff
    out = offL.copy()
    out[m] = np.Infinity
    return out

def mismatchFiltering(dL, dR, area=50, tau=1):
    '''
    Applies left-right and speckle filter
    Args:
        dL,dR: are numpy arrays containing the left and right disparity maps
        area: is the minimum area parameter of the speckle filter
        tau: maximum left-right disparity difference
    Returns:
        numpy array containing a filtered version of the left disparity map,
        where rejected pixels are set to infinity
    '''
    dLR = leftright(dL, dR, tau)
    dLRspeckle = specklefilter(dLR, area=area, th=1)
    return dLRspeckle
