# generic imports
import numpy as np
import os
import time
from psm import *
from hsm import *
#from deeppruner import *
from PIL import Image


# torch imports
import torch
from torchvision import transforms as T
import torch.nn.functional as F

# custom imports
import opt
import dataloader
import rectification_utils as rct_utils

import stereo_utils

import warnings
warnings.filterwarnings("ignore")

def test_inference(model, imgL, imgR):
    model.eval()
    with torch.no_grad():
        output = model(imgL, imgR)
    return output

def normalize_image(x, vmin=None, vmax=None):
    # input image is expected to be a HxWxC numpy array
    mi = np.min(x) if vmin is None else vmin
    ma = np.max(x) if vmax is None else vmax
    x = (x-mi)/(ma-mi+1e-8) # normalize to 0~1
    return x, mi, ma

def rescale_HWC_image(img, rescale_factor, method=Image.BICUBIC):
    # input image is expected to be a HxWxC numpy array
    # rescale_factor = 2 means upsample x2, rescale_factor = 0.5 means downsample x2
    h, w, _ = img.shape
    out_h, out_w = int(h * rescale_factor), int(w * rescale_factor)
    img_ = np.transpose(img, (2, 0, 1))
    img_ = T.Resize(size=(out_h, out_w), interpolation=method)(torch.Tensor(img_))
    img_ = np.transpose(img_.numpy(), (1, 2, 0))
    return img_

def load_pretrained_model(args):
    print(f"Loading {args.model} from {args.model_path} ...")

    t0 = time.time()
    if args.model == 'PSM':
        model = stackhourglass(args.max_disp)
    elif args.model == 'HSM':
        args.hsm_level = args.hsm_level if args.hsm_level in [1, 2, 3] else 1
        if os.path.basename(args.model_path) == 'vanilla_HSM.tar':
            args.max_disp = 768 # (192*4)
        model = hsm(args.max_disp, level=args.hsm_level, cuda=args.cuda)
    elif args.model == 'DeepPruner':
        model = deep_pruner()
        model.mode = 'evaluation'
    else:
        raise ValueError('Input model is not valid')

    # load pretrained model
    state_dict = torch.load(args.model_path, map_location='cpu')
    load_strict = True
    if args.model == "HSM" and args.hsm_level != 1:
        state_dict['state_dict'] = {k: v for k, v in state_dict['state_dict'].items() if 'disp' not in k}
        load_strict = False
    if args.model == "HSM" and os.path.basename(args.model_path) == 'vanilla_HSM.tar':
        state_dict['state_dict'] = {k: v for k, v in state_dict['state_dict'].items() if 'preconvs' not in k}

    # remove module. since the pretrained models were saved with nn.DataParallel
    state_dict_ = {k[7:]: v for k, v in state_dict['state_dict'].items() if k[:7] == "module."}
    model.load_state_dict(state_dict_, strict=load_strict)
    if args.cuda:
        model.to('cuda')
    else:
        model.to('cpu')

    print("...done in {:.2f} seconds".format(time.time() - t0))
    print('Number of model parameters: {}'.format(sum([p.data.nelement() for p in model.parameters()])))

    return model

def run_model(imgL_o, imgR_o, model, args, scaling_factor):
    normal_mean_var = {'mean': [0.485, 0.456, 0.406], 'std': [0.229, 0.224, 0.225]}
    infer_transform = T.Compose([T.ToTensor(), T.Normalize(**normal_mean_var)])

    # if grayscale input, make it 3-band like rgb input
    if imgL_o.shape[2] == 1:
        imgL_o = np.dstack([imgL_o, imgL_o, imgL_o])
    if imgR_o.shape[2] == 1:
        imgR_o = np.dstack([imgR_o, imgR_o, imgR_o])
    # clip negative values and set NaNs to 0
    nan_mask_left = np.isnan(imgL_o)[: , :, 0]
    nan_mask_right = np.isnan(imgR_o)[: , :, 0]
    imgL_o = imgL_o.clip(min=0)
    imgR_o = imgR_o.clip(min=0)
    imgL_o[np.isnan(imgL_o)] = 0.
    imgR_o[np.isnan(imgR_o)] = 0.

    # optional rescaling step
    if scaling_factor != 1:
        print("    Image rescaling of factor {}...".format(scaling_factor))
        start_time = time.time()
        imgL_o = rescale_HWC_image(imgL_o, scaling_factor) / 255.
        imgR_o = rescale_HWC_image(imgR_o, scaling_factor) / 255.
        nan_mask_left = np.dstack([nan_mask_left, nan_mask_left, nan_mask_left])
        nan_mask_right = np.dstack([nan_mask_right, nan_mask_right, nan_mask_right])
        nan_mask_left = rescale_HWC_image(nan_mask_left, scaling_factor, method=Image.NEAREST).astype(bool)
        nan_mask_right = rescale_HWC_image(nan_mask_right, scaling_factor, method=Image.NEAREST).astype(bool)
        nan_mask_left, nan_mask_right = nan_mask_left[:, :, 0], nan_mask_right[:, :, 0]
        print("    ...done in {:.2f} seconds".format(time.time() - start_time))

    # normalize input images
    mi, ma = min(imgR_o.min(), imgL_o.min()), max(imgR_o.max(), imgL_o.max())
    imgR_o = normalize_image(imgR_o, vmin=mi, vmax=ma)[0]
    imgL_o = normalize_image(imgL_o, vmin=mi, vmax=ma)[0]

    # convert to tensors
    imgR = infer_transform(imgR_o)
    imgL = infer_transform(imgL_o)

    # pad width and hight to make input dims multiple of D
    # input tensors are expected to have shape CxHxW
    if args.model == "PSM":
        D = 16
    elif args.model == "HSM":
        D = 64
    elif args.model == "DeepPruner":
        D = model.scale * 8
    else:
        raise ValueError("model is not valid")
    top_pad = int((imgL.shape[1] // D + 1) * D) - imgL.shape[1]
    right_pad = int((imgL.shape[2] // D + 1) * D) - imgL.shape[2]
    imgL = F.pad(imgL, (0, right_pad, top_pad, 0)).unsqueeze(0)
    imgR = F.pad(imgR, (0, right_pad, top_pad, 0)).unsqueeze(0)

    # run model
    start_time = time.time()
    if args.cuda:
        imgL, imgR = imgL.cuda(), imgR.cuda()
    print(f"    Running {args.model} model...")
    output = test_inference(model, imgL, imgR)
    if args.model == "HSM":
        output, entropy = output
    output = torch.squeeze(output).data.cpu().numpy()
    print("    ...done in {:.2f} seconds".format(time.time() - start_time))

    # undo padding
    pred_disp = output[top_pad:, :-right_pad or None]

    # rewrite nan values
    pred_disp[nan_mask_left] = np.nan
    imgL_o[nan_mask_left, :] = np.nan
    imgR_o[nan_mask_right, :] = np.nan

    return pred_disp, imgL_o, imgR_o

def save_output_files(pred_disp, imgL_o, imgR_o, pair_id, out_dir, scaling_factor=1, undo_scaling=True):
    # save input images and output disp as tif files
    
    for img, label in zip([imgL_o, imgR_o, pred_disp], ["rectified_left", "rectified_right", "disparity"]):
    
        # (0) undo scaling
        nan_mask = np.isnan(img)
        img[nan_mask] = np.inf
        if scaling_factor != 1. and undo_scaling:
            if label == 'disparity':
                img = rescale_HWC_image(img[..., np.newaxis], 1 / scaling_factor)[:, :, 0]
                img /= scaling_factor
            else:
                img = rescale_HWC_image(img, 1 / scaling_factor)
        img[np.isinf(img)] = np.nan
        
        # (1) save tiff    
        os.makedirs(os.path.join(out_dir, label), exist_ok=True)
        rct_utils.save_tif(os.path.join(out_dir, f"{label}/{pair_id}.tif"), img)

        # (2) extra: save all as png for web visualisation
        os.makedirs(os.path.join(out_dir, "png_" + label), exist_ok=True)
        nan_mask = np.isnan(img)
        img[nan_mask] = np.mean(img[~nan_mask])
        if label == 'disparity':
            png_img = normalize_image(np.dstack([img, img, img]))[0] * 255
            png_img[nan_mask, :] = 0
        else:
            png_img = normalize_image(img)[0] * 255
            png_img[nan_mask] = 0
        rct_utils.save_png(os.path.join(out_dir, f"png_{label}/{pair_id}.png"), png_img)


def test():

    # read input args
    args = opt.get_opt()

    # activate cuda only if available and desired
    args.cuda = args.cuda and torch.cuda.is_available()

    # load pretrained model
    model = load_pretrained_model(args)

    # load input data
    imgL_paths, imgR_paths = dataloader.dataloader(args.left_path, args.right_path)
    n_pairs = len(imgL_paths)
    print(f"Found {n_pairs} input stereo pairs\n")
    os.makedirs(args.out_dir, exist_ok=True)

    for pair_idx, (ref_path, sec_path) in enumerate(zip(imgL_paths, imgR_paths)):

        print("--> Pair {}/{}".format(pair_idx + 1, n_pairs))

        t0 = time.time()
        pair_id = rct_utils.get_file_id(ref_path)

        if args.invert_pair:
            ref_path_old = ref_path
            ref_path = sec_path
            sec_path = ref_path_old

        # optional rectification step
        if args.rectify:
            if args.roi is None:
                imgL_o = rct_utils.read_input_image(ref_path)
                x, y, w, h = 0, 0, imgL_o.shape[1], imgL_o.shape[0]
            else:
                x, y, w, h = args.roi
            print("    Rectifying...")
            start_time = time.time()
            rct_utils.rectify_pair_custom(ref_path, sec_path, x, y, w, h, args.out_dir)
            ref_path = os.path.join(args.out_dir, f"rectified_left/{pair_id}.tif")
            sec_path = os.path.join(args.out_dir, f"rectified_right/{pair_id}.tif")
            print("    ...done in {:.2f} seconds".format(time.time() - start_time))

        imgL_o = rct_utils.read_input_image(ref_path)
        imgR_o = rct_utils.read_input_image(sec_path)

        pred_disp, imgL_o, imgR_o = run_model(imgL_o, imgR_o, model, args, scaling_factor=args.img_scaling)
        save_output_files(pred_disp, imgL_o, imgR_o, pair_id, args.out_dir, args.img_scaling, args.undo_scaling)

        if args.lr_check:
            args.out_dir = os.path.join(args.out_dir, "lr_check")
            os.makedirs(args.out_dir, exist_ok=True)
            imgL_o_flip = np.fliplr(imgL_o)
            imgR_o_flip = np.fliplr(imgR_o)
            pred_disp_flip, imgR_o_flip, imgL_o_flip = run_model(imgR_o_flip, imgL_o_flip, model, args, scaling_factor=1)
            pred_disp_flip = np.fliplr(pred_disp_flip).copy()
            save_output_files(pred_disp_flip, imgL_o_flip, imgR_o_flip, pair_id, args.out_dir, args.img_scaling, args.undo_scaling)
            #filtered_disp = stereo_utils.leftright(pred_disp, pred_disp_flip, maxdiff=2)
            print("running left right check with threshold = {:.2f}".format(args.lr_thr))
            filtered_disp = stereo_utils.mismatchFiltering(pred_disp, pred_disp_flip, area=20, tau=args.lr_thr)
            print("done")
            os.makedirs(os.path.join(args.out_dir, "filt_disp"), exist_ok=True)
            if args.img_scaling != 1. and args.undo_scaling:
                filtered_disp = rescale_HWC_image(filtered_disp[..., np.newaxis], 1 / args.img_scaling)[:, :, 0]
                filtered_disp /= args.img_scaling
            filtered_disp[np.isinf(filtered_disp)] = np.nan
            rct_utils.save_tif(os.path.join(args.out_dir, f"filt_disp/{pair_id}.tif"), filtered_disp)

        print("    Total time: {:.2f} seconds\n".format(time.time() - t0))

if __name__ == '__main__':
    test()
